'''
# Copy number plotting

Plotting functions for displaying copy number taking tables of copy number predictions as input.

'''

import math
import itertools
import collections
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import colorConverter
import seaborn

import wgs_analysis.plots.colors
import wgs_analysis.plots.positions
import wgs_analysis.plots.utils
import wgs_analysis.algorithms as algorithms
import wgs_analysis.algorithms.cnv
import wgs_analysis.refgenome
import wgs_analysis.algorithms.merge


def create_segments(df, field):
    segments = np.array([[df['start'].values, df[field].values], [df['end'].values, df[field].values]])
    segments = np.transpose(segments, (2, 0, 1))
    return segments


def create_connectors(df, field):
    prev = df.iloc[:-1].reset_index()
    next = df.iloc[1:].reset_index()
    mids = ((prev[field] + next[field]) / 2.0).values
    prev_cnct = np.array([[prev['end'].values, prev[field].values], [prev['end'].values, mids]])
    prev_cnct = np.transpose(prev_cnct, (2, 0, 1))
    next_cnct = np.array([[next['start'].values, mids], [next['start'].values, next[field].values]])
    next_cnct = np.transpose(next_cnct, (2, 0, 1))
    return np.concatenate([prev_cnct, next_cnct])


def create_quads(df, field):
    quads = np.array([
        [df['start'].values, np.zeros(len(df.index))],
        [df['start'].values, df[field].values],
        [df['end'].values, df[field].values],
        [df['end'].values, np.zeros(len(df.index))],
    ])
    quads = np.transpose(quads, (2, 0, 1))
    return quads


def plot_segments(ax, cnv, value_col, color, fill=False, fill_alpha=0.5, lw=1):
    """
    Plot segment copy number as line plots

    Args:
        ax (matplotlib.axes.Axes): plot axes
        cnv (pandas.DataFrame): cnv table
        value_col (str): column name to plot
        color (str or tuple): color of lines

    Plot segment copy number as line plots.  The columns 'start' and 'end'
    are expected and should be adjusted for full genome plots.  Values from
    the column given by 'value_col' are plotted with color given by 'color'.

    """ 

    cnv = cnv.sort_values('start')

    segments = create_segments(cnv, value_col)
    ax.add_collection(matplotlib.collections.LineCollection(segments, colors=color, lw=lw))

    connectors = create_connectors(cnv, value_col)
    ax.add_collection(matplotlib.collections.LineCollection(connectors, colors=color, lw=lw))

    if fill:
        quad_color = colorConverter.to_rgba(color, alpha=fill_alpha)
        quads = create_quads(cnv, value_col)
        ax.add_collection(matplotlib.collections.PolyCollection(quads, facecolors=quad_color, edgecolors=quad_color, lw=0))


def plot_cnv_segments(ax, cnv, column, segment_color, fill=False):
    """ Plot raw copy number as line plots

    Args:
        ax (matplotlib.axes.Axes): plot axes
        cnv (pandas.DataFrame): cnv table
        column (str): name of copies column
        segment_color: color of the segments

    Plot copy number as line plots.  The columns 'start' and 'end'
    are expected and should be adjusted for full genome plots.  Values from the
    'column' columns are plotted.

    """ 

    quad_color = colorConverter.to_rgba(segment_color, alpha=0.5)

    cnv = cnv.sort_values('start')

    segments = create_segments(cnv, column)
    ax.add_collection(matplotlib.collections.LineCollection(segments, colors=segment_color, lw=1))

    connectors = create_connectors(cnv, column)
    ax.add_collection(matplotlib.collections.LineCollection(connectors, colors=segment_color, lw=1))

    if fill:
        quads = create_quads(cnv, column)
        ax.add_collection(matplotlib.collections.PolyCollection(quads, facecolors=quad_color, edgecolors=quad_color, lw=0))


def plot_cnv_genome(ax, cnv, maxcopies=4, minlength=1000, major_col='major_raw', minor_col='minor_raw', scatter=False, squashy=False):
    """
    Plot major/minor copy number across the genome

    Args:
        ax (matplotlib.axes.Axes): plot axes
        cnv (pandas.DataFrame): `cnv_site` table
        maxcopies (int): maximum number of copies for setting y limits
        minlength (int): minimum length of segments to be drawn
        major_col (str): name of column to use as major copy number
        minor_col (str): name of column to use as minor copy number
        scatter (boolean): display segments as scatter points not segments
        squashy (boolean): squash the y axis to display all copy numbers

    """

    segment_color_major = plt.get_cmap('RdBu')(0.1)
    segment_color_minor = plt.get_cmap('RdBu')(0.9)

    cnv = cnv.copy()

    squash_coeff = 0.15
    squash_f = lambda a: np.tanh(squash_coeff * a)
    if squashy:
        cnv[major_col] = squash_f(cnv[major_col])
        cnv[minor_col] = squash_f(cnv[minor_col])

    if 'length' not in cnv:
        cnv['length'] = cnv['end'] - cnv['start']

    cnv = cnv[['chromosome', 'start', 'end', 'length', major_col, minor_col]]

    cnv = cnv[cnv['length'] >= minlength]
    
    cnv = cnv[cnv['chromosome'].isin(wgs_analysis.refgenome.info.chromosomes)]

    cnv.set_index('chromosome', inplace=True)
    cnv['chromosome_start'] = wgs_analysis.refgenome.info.chromosome_start
    cnv.reset_index(inplace=True)

    cnv['start'] = cnv['start'] + cnv['chromosome_start']
    cnv['end'] = cnv['end'] + cnv['chromosome_start']

    if scatter:
        cnv['mid'] = 0.5 * (cnv['start'] + cnv['end'])
        for column, color in ((minor_col, segment_color_minor), (major_col, segment_color_major)):
            clipped_cnv = cnv[cnv[column] < maxcopies]
            amp_cnv = cnv[cnv[column] >= maxcopies]
            ax.scatter(clipped_cnv['mid'], clipped_cnv[column], color=color, s=2, alpha=1)
            ax.scatter(clipped_cnv['mid'], clipped_cnv[column], color=color, s=10, alpha=0.1)
            ax.scatter(amp_cnv['mid'], np.ones(amp_cnv.shape[0]) * maxcopies, color=color, s=30)

    else:
        plot_cnv_segments(ax, cnv, column=major_col, segment_color=segment_color_major)
        plot_cnv_segments(ax, cnv, column=minor_col, segment_color=segment_color_minor)

    ax.spines['left'].set_position(('outward', 5))
    ax.spines['bottom'].set_position(('outward', 5))
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    ax.set_xlim((-0.5, wgs_analysis.refgenome.info.chromosome_end.max()))
    ax.set_xlabel('chromosome')
    ax.set_xticks([0] + list(wgs_analysis.refgenome.info.chromosome_end.values))
    ax.set_xticklabels([])

    if squashy:
        yticks = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 20])
        yticks_squashed = squash_f(yticks)
        ytick_labels = [str(a) for a in yticks]
        ax.set_yticks(yticks_squashed)
        ax.set_yticklabels(ytick_labels)
        ax.set_ylim((-0.01, 1.01))
        ax.spines['left'].set_bounds(0, 1)
    else:
        ax.set_ylim((-0.05*maxcopies, maxcopies))
        ax.set_yticks(range(0, int(maxcopies) + 1))
        ax.spines['left'].set_bounds(0, maxcopies)

    ax.xaxis.tick_bottom()
    ax.yaxis.tick_left()
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(wgs_analysis.refgenome.info.chromosome_mid))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(wgs_analysis.refgenome.info.chromosomes))

    ax.xaxis.grid(True, which='major', linestyle=':')
    ax.yaxis.grid(True, which='major', linestyle=':')


def plot_cnv_chromosome(ax, cnv, chromosome, start=None, end=None, maxcopies=4, minlength=1000, major_col='major_raw', minor_col='minor_raw', fontsize=None, scatter=False, squashy=False, squash_coeff=None, grid=True, yticks=None):
    """
    Plot major/minor copy number across a chromosome

    Args:
        ax (matplotlib.axes.Axes): plot axes
        cnv (pandas.DataFrame): `cnv_site` table
        chromosome (str): chromosome to plot
        start (int): start of range in chromosome to plot
        end (int): start of range in chromosome to plot
        maxcopies(float): maximum number of copies for setting y limits
        minlength(int): minimum length of segments to be drawn
        major_col (str): name of column to use as major copy number
        minor_col (str): name of column to use as minor copy number
        fontsize (int): size of x tick labels
        scatter (boolean): display segments as scatter points not segments
        squashy (boolean): squash the y axis to display all copy numbers
        squash_coeff (float): coefficient for squashing y axis
        grid (boolean): whether to display grid on plot
        yticks (array): override y ticks

    """

    chromosome_length = wgs_analysis.refgenome.info.chromosome_info.set_index('chr').loc[chromosome, 'chromosome_length']

    segment_color_major = plt.get_cmap('RdBu')(0.1)
    segment_color_minor = plt.get_cmap('RdBu')(0.9)

    cnv = cnv.copy()

    if squash_coeff is None:
        squash_coeff = 0.15
    squash_f = lambda a: np.tanh(squash_coeff * a)
    if squashy:
        cnv[major_col] = squash_f(cnv[major_col])
        cnv[minor_col] = squash_f(cnv[minor_col])

    cnv = cnv.loc[(cnv['chromosome'] == chromosome)]
    cnv = cnv.loc[(cnv['length'] >= minlength)]

    if start is None:
        start = 0
    if end is None:
        end = chromosome_length

    cnv['genomic_length'] = cnv['end'] - cnv['start']
    cnv['length_fraction'] = cnv['length'].astype(float) / cnv['genomic_length'].astype(float)
    cnv = cnv.loc[(cnv['length_fraction'] >= 0.5)]

    cnv = cnv.loc[(cnv['end'] > start)]
    cnv = cnv.loc[(cnv['start'] < end)]
    
    if scatter:
        cnv['mid'] = 0.5 * (cnv['start'] + cnv['end'])
        for column, color in ((minor_col, segment_color_minor), (major_col, segment_color_major)):
            clipped_cnv = cnv[cnv[column] < maxcopies]
            amp_cnv = cnv[cnv[column] >= maxcopies]
            ax.scatter(clipped_cnv['mid'], clipped_cnv[column], color=color, s=2, alpha=1)
            ax.scatter(clipped_cnv['mid'], clipped_cnv[column], color=color, s=10, alpha=0.1)
            ax.scatter(amp_cnv['mid'], np.ones(amp_cnv.shape[0]) * maxcopies, color=color, s=30)

    else:
        plot_cnv_segments(ax, cnv, column=major_col, segment_color=segment_color_major)
        plot_cnv_segments(ax, cnv, column=minor_col, segment_color=segment_color_minor)

    ax.spines['left'].set_position(('outward', 5))
    ax.spines['bottom'].set_position(('outward', 5))
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    xticks = np.arange(0, chromosome_length, 2e7)
    xticklabels = ['{0:d}M'.format(int(x / 1e6)) for x in xticks]
    xminorticks = np.arange(0, chromosome_length, 1e6)
    ax.set_xlabel(f'chromosome {chromosome}')
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticklabels)
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(xminorticks))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
    ax.set_xlim((start, end))

    if squashy:
        if yticks is None:
            yticks = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 10, 20])
        yticks_squashed = squash_f(np.array(yticks))
        ytick_labels = [str(a) for a in yticks]
        ax.set_yticks(yticks_squashed)
        ax.set_yticklabels(ytick_labels)
        ax.set_ylim((-0.01, 1.01))
        ax.spines['left'].set_bounds(0, 1)
    else:
        if yticks is None:
            yticks = range(0, int(maxcopies) + 1, 2)
        ax.set_ylim((-0.05*maxcopies, maxcopies))
        ax.set_yticks(yticks)
        ax.spines['left'].set_bounds(0, maxcopies)
  
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    wgs_analysis.plots.utils.trim_spines_to_ticks(ax)

    if grid:
        ax.xaxis.grid(True, which='major', linestyle=':')
        ax.yaxis.grid(True, which='major', linestyle=':')


def plot_cnv(cnv, site_ids, chromosome, start=None, end=None, maxcopies=2):
    """
    Plot major/minor copy number across a chromosome and breakends as flags

    Args:
        cnv (pandas.DataFrame): `cnv_site` table
        site_ids (str): sites to plot
        chromosome (str): chromosome to plot
        start (int): start of range in chromosome to plot
        end (int): start of range in chromosome to plot
        maxcopies(int): maximum number of copies for setting y limits

    """
    
    site_axes = collections.OrderedDict()
    
    num_sites = len(site_ids)

    for fig_idx, sample_id in enumerate(site_ids):
        
        ax = plots.utils.setup_axes(plt.subplot(num_sites, 1, fig_idx+1))
        
        plots.cnv.plot_cnv_chromosome(ax, cnv, sample_id, chromosome, start=start, end=end, maxcopies=maxcopies)

        if fig_idx != num_sites - 1:
            ax.set_xticklabels([])
            ax.set_xlabel('')
            ax.spines['bottom'].set_visible(False)
            ax.xaxis.set_ticks_position('none')

        site_axes[sample_id] = ax
        
    plt.tight_layout()
    
    return site_axes


def plot_cnv_brks(cnv, brks, site_ids, chromosome, start=None, end=None, maxcopies=2):
    """
    Plot major/minor copy number across a chromosome and breakends as flags

    Args:
        cnv (pandas.DataFrame): `cnv_site` table
        brks (pandas.DataFrame): `breakend_site` table
        site_ids (str): sites to plot
        chromosome (str): chromosome to plot
        start (int): start of range in chromosome to plot
        end (int): start of range in chromosome to plot
        maxcopies(int): maximum number of copies for setting y limits

    """
    
    site_axes = collections.OrderedDict()
    
    num_sites = len(site_ids)

    for fig_idx, sample_id in enumerate(site_ids):
        
        site_brks = brks[brks['sample_id'] == sample_id]
        
        ax = plots.utils.setup_axes(plt.subplot(num_sites, 1, fig_idx+1))
        
        plots.cnv.plot_cnv_chromosome(ax, cnv, sample_id, chromosome, start=start, end=end, maxcopies=maxcopies)
        plots.positions.plot_breakends(ax, brks, sample_id, chromosome, start=start, end=end)

        if fig_idx != num_sites - 1:
            ax.set_xticklabels([])
            ax.set_xlabel('')
            ax.spines['bottom'].set_visible(False)
            ax.xaxis.set_ticks_position('none')

        site_axes[sample_id] = ax
        
    plt.tight_layout()
    
    return site_axes


def plot_cnv_scatter(ax, cnv, site, chromosome_color, point_alpha=1.0, line_alpha=1.0):

    cnv = cnv[cnv['sample_id'] == site]

    cnv = cnv[cnv['length'] >= 10000]
    
    cnv = cnv.set_index(['chromosome', 'start', 'end'])

    cs = [chromosome_color[c[0]] for c in cnv.index.values]
    sz = cnv['length'] / 100000.

    points = ax.scatter(cnv['major_raw'], cnv['minor_raw'], s=sz, facecolor=cs, edgecolor=cs, linewidth=0.0, zorder=2, alpha=point_alpha)

    ax.set_xlim((-0.5, cnv['major_raw'].quantile(0.95) + 0.5))
    ax.set_ylim((-0.5, cnv['minor_raw'].quantile(0.95) + 0.5))


def plot_cnv_scatter_pairwise(ax, cnv, site_a, site_b, chromosome_color, point_alpha=1.0, line_alpha=1.0):

    cnv = cnv.set_index(['chromosome', 'start', 'end', 'sample_id']).unstack()
    cnv = cnv.fillna(0.0)

    cs = [chromosome_color[c[0]] for c in cnv.index.values]
    sz = (cnv['length'][site_a] + cnv['length'][site_b])/500000.

    for x1, y1, x2, y2, c, s in zip(cnv['major_raw'][site_a], cnv['minor_raw'][site_a], cnv['major_raw'][site_b], cnv['minor_raw'][site_b], cs, sz):
        lines = plt.plot([x1, x2], [y1, y2], alpha=line_alpha, zorder=1, color=c, lw=1.0)

    points1 = ax.scatter(cnv['major_raw'][site_a], cnv['minor_raw'][site_a], s=sz, facecolor=cs, edgecolor=cs, linewidth=0.0, zorder=2, alpha=point_alpha)
    points2 = ax.scatter(cnv['major_raw'][site_b], cnv['minor_raw'][site_b], s=sz, facecolor=cs, edgecolor=cs, linewidth=0.0, zorder=2, alpha=point_alpha)


def plot_patient_cnv(cnv):

    sites = cnv['sample_id'].unique()

    plots.utils.setup_plot()
    fig = plt.figure(figsize=(12,5*len(sites)))

    chromosome_color = plots.colors.create_chromosome_color_map()

    for idx, sample_id in enumerate(sites):
        ax = plots.utils.setup_axes(plt.subplot(len(sites), 1, idx+1))
        plot_cnv_scatter(ax, cnv, sample_id, chromosome_color)
        if idx == len(sites) - 1:
            ax.set_xlabel('major copy number')
            ax.set_ylabel('minor copy number')
        ax.set_title(helpers.plot_ids[sample_id] + ' copies')

    fig.legend([plt.Circle((0, 0), radius=1, color=color) for chrom, color in chromosome_color.items()], chromosome_color.keys(), loc="lower right")

    plt.tight_layout()

    return fig


def plot_cnv_pairwise(ax, cnv, site_a, site_b, chromosome_color, highlighted_chromosomes=None):

    cnv = cnv[(cnv['sample_id'] == site_a) | (cnv['sample_id'] == site_b)]
    
    cnv = algorithms.cnv.cluster_copies(cnv)
    
    cnv = cnv[cnv['length'] >= 5000000]
    
    if highlighted_chromosomes is None:
        plot_cnv_scatter_pairwise(ax, cnv, site_a, site_b, chromosome_color, 1.0, 0.5)
    else:
        plot_cnv_scatter_pairwise(ax, cnv[~(cnv['chromosome'].isin(highlighted_chromosomes))], site_a, site_b, chromosome_color, 0.1, 0.1)
        plot_cnv_scatter_pairwise(ax, cnv[cnv['chromosome'].isin(highlighted_chromosomes)], site_a, site_b, chromosome_color, 1.0, 1.0)

    ax.grid(True)


def plot_patient_cnv_pairwise(fig, cnv, highlighted_chromosomes=None):

    pairs = list(itertools.combinations(cnv['sample_id'].unique(), 2))

    chromosome_color = plots.colors.create_chromosome_color_map()

    subplot_width = np.ceil(np.sqrt(len(pairs)))
    subplot_height = np.ceil(len(pairs) / subplot_width)

    for idx, (site_a, site_b) in enumerate(pairs):
        ax = plots.utils.setup_axes(plt.subplot(subplot_height, subplot_width, idx+1))
        plot_cnv_pairwise(ax, cnv, site_a, site_b, chromosome_color, highlighted_chromosomes)
        if idx == subplot_height * (subplot_width - 1):
            ax.set_xlabel('major copy number')
            ax.set_ylabel('minor copy number')
        ax.set_title(helpers.plot_ids[site_a] + ' vs ' + helpers.plot_ids[site_b])

    fig.legend([plt.Circle((0, 0), radius=1, color=color) for chrom, color in chromosome_color.items()], chromosome_color.keys(), loc="lower right")

    plt.tight_layout()


def create_uniform_segments(segment_length, max_end):
    """
    Create a table of uniform segments with a given segment length

    Args:
        segment_length (int): uniform segment length
        max_end (int): maximum end position of segments

    Returns:
        pandas.DataFrame: table of uniform segments

    The returned table will have columns start, end.  Segments are created according
    to given max end.
    """

    num_segments = np.ceil(float(max_end) / float(segment_length))

    starts = np.arange(0, num_segments*segment_length, segment_length)
    starts = starts.astype(int) + 1

    segments = pd.DataFrame({'start':starts})
    segments['end'] = segments['start'] + segment_length - 1

    return segments


def create_uniform_segments_genome(segment_length):
    """
    Create a table of uniform segments with a given segment length

    Args:
        segment_length (int): uniform segment length

    Returns:
        pandas.DataFrame: table of uniform segments

    The returned table will have columns chromosome, start, end.  Segments are created according
    to chromosome lengths.
    """

    num_segments = (wgs_analysis.refgenome.info.chromosome_lengths.astype(float) / segment_length).apply(np.ceil)

    chroms = np.concatenate([np.repeat(c, n) for c, n in num_segments.iteritems()])
    starts = np.concatenate([np.arange(0, m*segment_length, segment_length) for c, m in num_segments.iteritems()])

    chroms = chroms.astype(str)
    starts = starts.astype(int) + 1

    segments = pd.DataFrame({'chromosome':chroms, 'start':starts})
    segments['end'] = segments['start'] + segment_length - 1

    return segments


def uniform_resegment(cnv, segment_length=100000):
    """
    Create a table of uniform segments data from arbitrary segments segment data

    Args:
        cnv (pandas.DataFrame): segment data
        segment_length (int): uniform segment length

    Returns:
        pandas.DataFrame: resegmented table

    The cnv table should have columns start, end.  Returns a resegmented
    table with columns start, end, start_reseg, end_reseg.  Original 
    rows will be represented multiple times if they are split by uniform segments,
    and can be identified by having the same start, end.  The start_reseg
    and end_reseg columns represent the subsegment resulting from the split.
    """

    # First segment cannot start at coordinate less than 1
    assert cnv['start'].min() >= 1

    cnv['seg_idx'] = xrange(len(cnv.index))

    uniform_segments = create_uniform_segments(segment_length, cnv['end'].max())
    uniform_segments['reseg_idx'] = xrange(len(uniform_segments.index))

    # Find starts of the resegmentation that fall within cnv segments
    seg_idx_1, reseg_idx_1 = wgs_analysis.algorithms.merge.interval_position_overlap_unsorted(
        cnv[['start', 'end']].values,
        uniform_segments['start'].values)
    reseg_1 = pd.DataFrame({'seg_idx': seg_idx_1, 'reseg_idx': reseg_idx_1})

    # Find ends of the resegmentation that fall within cnv segments
    seg_idx_2, reseg_idx_2 = wgs_analysis.algorithms.merge.interval_position_overlap_unsorted(
        cnv[['start', 'end']].values,
        uniform_segments['end'].values)
    reseg_2 = pd.DataFrame({'seg_idx': seg_idx_2, 'reseg_idx': reseg_idx_2})

    # Find starts of the cnv segments that fall within the resegmentation
    reseg_idx_3, seg_idx_3 = wgs_analysis.algorithms.merge.interval_position_overlap_unsorted(
        uniform_segments[['start', 'end']].values,
        cnv['start'].values)
    reseg_3 = pd.DataFrame({'seg_idx': seg_idx_3, 'reseg_idx': reseg_idx_3})

    reseg = pd.concat([reseg_1, reseg_2, reseg_3], ignore_index=True).drop_duplicates()

    cnv = cnv.merge(reseg, on='seg_idx')
    cnv = cnv.merge(uniform_segments, on='reseg_idx', suffixes=('', '_reseg'))

    assert np.all(
        ((cnv['start'] >= cnv['start_reseg']) & (cnv['start'] <= cnv['end_reseg'])) |
        ((cnv['end'] >= cnv['start_reseg']) & (cnv['end'] <= cnv['end_reseg'])) |
        ((cnv['start'] <= cnv['start_reseg']) & (cnv['end'] >= cnv['end_reseg'])))

    return cnv


def uniform_resegment_genome(cnv, segment_length=100000):
    """
    Create a table of uniform segments data from arbitrary segments segment data

    Args:
        cnv (pandas.DataFrame): segment data
        segment_length (int): uniform segment length

    Returns:
        pandas.DataFrame: resegmented table

    The cnv table should have columns chromosome, start, end.  Returns a resegmented
    table with columns chromosome, start, end, start_reseg, end_reseg.  Original 
    rows will be represented multiple times if they are split by uniform segments,
    and can be identified by having the same chromosome, start, end.  The start_reseg
    and end_reseg columns represent the subsegment resulting from the split.
    """

    return cnv.groupby('chromosome').apply(lambda cnv: uniform_resegment(cnv, segment_length=100000))


def uniform_segment_copies(cnv, sample_column, data_columns, segment_length=100000):
    """
    Create a table of uniformly segmented data from arbitrarily segment data

    Args:
        cnv (pandas.DataFrame): segment data
        sample_column (list): sample id column
        data_columns (list): data columns to resegment
        segment_length (int): uniform segment length

    Returns:
        pandas.DataFrame: resegmented table

    The cnv table should have sample_column, chrom, start and end columns in addition to
    the columns for which resegmentation is requested.  Returns a resegmented table
    with sample_column, chrom, segment_start, segment_end columns, in addition to the 
    requested columns calculated as length weighted averages of the original values.
    """

    cnv_reseg = uniform_resegment_genome(cnv[['chromosome', 'start', 'end']], segment_length=segment_length)

    cnv_reseg = cnv_reseg.merge(cnv, on=['chromosome', 'start', 'end'])

    cnv_reseg['segment_start'] = cnv_reseg['start_reseg'] / segment_length
    cnv_reseg['segment_start'] = cnv_reseg['segment_start'].astype(int) * segment_length + 1

    cnv_reseg.set_index([sample_column, 'chromosome', 'segment_start'], inplace=True)

    # Average requested columns weighted by length of segment
    # exclude null values in calculation
    for column in data_columns:

        # Length of resegmented segments
        cnv_reseg['length_reseg'] = cnv_reseg['end_reseg'] - cnv_reseg['start_reseg'] + 1

        # Mask segments with null values from normalization calculation
        cnv_reseg['length_reseg'] *= (cnv_reseg[column].notnull() * 1)

        # Normalize by total length of resegmented segments
        cnv_reseg['length_total_reseg'] = cnv_reseg.groupby(level=[0, 1, 2])['length_reseg'].sum()
        cnv_reseg['weight_reseg'] = (
            cnv_reseg['length_reseg'].astype(float) /
            cnv_reseg['length_total_reseg'].astype(float))
        cnv_reseg[column] *= cnv_reseg['weight_reseg']

        # Mask segments with null values from summation
        cnv_reseg[column] = cnv_reseg[column].fillna(0.0)

    cnv_reseg = cnv_reseg.groupby(level=[0, 1, 2])[data_columns].sum()

    cnv_reseg.reset_index(inplace=True)

    # Ensure the segments are consistent regardless of the cnv data
    seg_full = (
        create_uniform_segments_genome(segment_length)
        .rename(columns={'start':'segment_start'})
        .set_index(['chromosome', 'segment_start']))

    cnv_reseg = (
        cnv_reseg
        .set_index(['chromosome', 'segment_start', sample_column])
        .unstack()
        .reindex(seg_full.index)
        .stack(dropna=False)
        .reset_index())

    cnv_reseg['segment_end'] = cnv_reseg['segment_start'] + segment_length - 1

    return cnv_reseg


def plot_loh(ax, segment_table, stats_table, segment_length=100000):

    index_cols = ['chromosome', 'segment_start', 'sample_id']

    minor = segment_table.set_index(index_cols)['minor_raw']\
                         .unstack()\
                         .fillna(1.0)
    
    loh = 1.0 - minor
    loh = np.maximum(loh, 0.0)

    plot_cn_matrix(ax, loh, segment_length=segment_length)


def plot_total(ax, segment_table, stats_table, segment_length=100000):

    index_cols = ['chromosome', 'segment_start', 'sample_id']

    segment_matrix = segment_table.set_index(index_cols)[['major_raw', 'minor_raw']].unstack()

    total = segment_matrix['major_raw'] + segment_matrix['minor_raw']

    ploidy = stats_table.set_index('sample_id')['ploidy']

    for sample_id in total.keys():
        total[sample_id] /= ploidy[sample_id]

    total = total.fillna(1.0)
    
    plot_cn_matrix(ax, total, segment_length=segment_length, norm_max=2.0)


def plot_subclonal(ax, segment_table, stats_table, segment_length=100000):

    index_cols = ['chromosome', 'segment_start', 'sample_id']

    subclonal = segment_table.set_index(index_cols)['subclonal']\
                             .unstack()\
                             .fillna(0.0)

    plot_cn_matrix(ax, subclonal, segment_length=segment_length)


def plot_cn_matrix(ax, matrix, segment_length=100000, norm_min=0.0, norm_max=1.0, cmap=None):
    
    if cmap is None:
        cmap = plt.get_cmap('Oranges')

    chromosome_index = [(chrom, idx) for idx, chrom in enumerate(wgs_analysis.refgenome.info.chromosomes)]
    chromosome_index = pd.DataFrame(chromosome_index, columns=['chromosome', 'chromosome_idx'])

    matrix = (
        matrix
        .reset_index()
        .merge(chromosome_index)
        .drop('chromosome', axis=1)
        .set_index(['chromosome_idx', 'start'])
        .sort_index()
    )

    ax.imshow(
        matrix.values.T, aspect='auto', 
        interpolation='nearest',
        cmap=cmap,
        norm=matplotlib.colors.Normalize(vmin=norm_min, vmax=norm_max, clip=True),
    )

    chrom_seg_cnt = (
        matrix.reset_index()
        .groupby('chromosome_idx')
        .size()
        .sort_index()
    )

    chrom_seg_end = chrom_seg_cnt.cumsum()
    chrom_seg_mid = chrom_seg_end - chrom_seg_cnt / 2.0

    ax.set_xlim((-0.5, len(matrix.index)))
    ax.set_xticks([0] + list(chrom_seg_end.values))
    ax.set_xticklabels([])
    ax.set_yticks(xrange(len(matrix.columns.values)))
    ax.set_yticklabels(matrix.columns.values)
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(chrom_seg_mid))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(wgs_analysis.refgenome.info.chromosomes))
    ax.xaxis.grid(True, which='major', color='white', linestyle='-', linewidth=2)
    ax.xaxis.grid(False, which='minor')
    ax.yaxis.grid(False, which='major')
    ax.yaxis.grid(False, which='minor')




'''
# Tree related tables

Generation of derivative tables for tree related questions.

'''

import numpy as np
import pandas as pd
import scipy
import dollo.run
import scipy.misc

import wgs_analysis.helpers as helpers
import wgs_analysis.params as params
import wgs_analysis.annotation as annotation
import wgs_analysis.annotation.cnv
import wgs_analysis.algorithms as algorithms
import wgs_analysis.algorithms.tree
import wgs_analysis.plots as plots
import wgs_analysis.plots.utils


def convergent_loh(tree, cnv):
    """ Calculate a table of loh that doesnt fit a tree.

    Args:
        tree (TreeNode): tree to test cnvs for convergence
        cnv (pandas.DataFrame): table of cnvs

    Returns:
        pandas.DataFrame: table of loh events with columns `length`, `is_loh`, `is_concordant`, `node_list`

    """

    cnv['length'] = cnv['end'] - cnv['start']
    cnv['length'] = cnv['length'].astype(float)

    cnv = annotation.cnv.annotate_cnv_type(cnv)

    loh_site_matrix = cnv.set_index(['chrom', 'start', 'end', 'length', 'site_id'])['loh'].unstack().dropna()

    def add_loh_state(tree, loh_sites):
        for child in tree.children:
            add_loh_state(child, loh_sites)
        if tree.is_leaf:
            tree.loh = (tree.name in loh_sites) * 1
        else:
            tree.loh = (all([child.loh for child in tree.children])) * 1

    def add_loh_origin(tree):
        if tree.loh:
            tree.loh_origin = 1
            for descendent in tree.descendents:
                descendent.loh_origin = 0
        else:
            tree.loh_origin = 0
            for child in tree.children:
                add_loh_origin(child)

    def calc_loh_origins(row):
        loh_sites = set(row[row > 0].index.values)
        add_loh_state(tree, loh_sites)
        add_loh_origin(tree)
        loh_nodes = pd.Series(tree.get_attr_map('loh_origin'))
        return loh_nodes
        
    loh_origin_matrix = loh_site_matrix.apply(calc_loh_origins, axis=1)

    nodes = list(loh_origin_matrix.columns.values)

    loh_origin_matrix = loh_origin_matrix.reset_index()

    loh_origin_matrix = loh_origin_matrix.set_index(nodes)

    loh_origin_lengths = loh_origin_matrix.groupby(level=range(len(nodes)))['length'].sum().reset_index()

    def format_node_list(row):
        return ','.join([str(a) for a in row[row > 0].index])

    loh_origin_lengths['is_loh'] = (loh_origin_lengths[nodes].sum(axis=1) >= 1) * 1
    loh_origin_lengths['is_concordant'] = (loh_origin_lengths[nodes].sum(axis=1) <= 1) * 1
    loh_origin_lengths['node_list'] = loh_origin_lengths[nodes].apply(format_node_list, axis=1)
    loh_origin_lengths = loh_origin_lengths[loh_origin_lengths['is_loh'] == 1]
    loh_origin_lengths = loh_origin_lengths.drop(nodes, axis=1)

    return loh_origin_lengths


def snvs_deleted_table(nodes):

    nodes['is_lost'] = nodes['ml_loss']
    nodes['is_deleted'] = (nodes['deletion'] == 1) & (nodes['is_lost'] == 1)

    data = nodes.groupby(['chrom', 'coord', 'ref', 'alt'])[['is_lost', 'is_deleted']]\
                .any()\
                .reset_index()

    data = data[data['is_lost']].groupby(['is_deleted'])\
                                .size()\
                                .reset_index()\
                                .rename(columns={0:'count'})

    data['is_deleted'] *= 1

    return data


def snvs_possibly_deleted_table():

    lost_table = list()

    for patient_id in helpers.get_wgss_patient_ids():

        tree = helpers.get_snv_tree(patient_id)
        nodes = helpers.get_snv_nodes(patient_id)

        siblings = list()
        for node in tree.nodes:
            if node.is_leaf:
                continue
            assert len(node.children) == 2
            child_labels = [child.label for child in node.children]
            siblings.append(tuple(child_labels))
            siblings.append(tuple(reversed(child_labels)))
        siblings = pd.DataFrame(siblings, columns=['node', 'sibling_node'])

        sibling_nodes = nodes[['chrom', 'coord', 'ref', 'alt', 'node', 'deletion']].copy()
        sibling_nodes = sibling_nodes.rename(columns={'node':'sibling_node', 'deletion':'sibling_deletion'})

        nodes = nodes.merge(siblings)

        nodes = nodes.merge(sibling_nodes)

        loss = nodes.loc[(nodes['ml_origin'] == 1) &
                         (nodes['sibling_deletion'] == 1),
                         ['chrom', 'coord', 'ref', 'alt', 'node', 'sibling_node']]

        lost_table.append((patient_id, len(loss.index)))

    lost_table = pd.DataFrame(lost_table, columns=('patient_id', 'loss_count'))

    return lost_table


def correlate_patient_origin_loss():

    event_table = list()

    for patient_id in helpers.get_wgss_patient_ids():

        snv_tree = helpers.get_snv_tree(patient_id)
        breakpoint_tree = helpers.get_breakpoint_tree(patient_id)

        try:
            snv_breakpoint_nodes = algorithms.tree.align_trees(snv_tree, breakpoint_tree)
        except ValueError:
            print 'patient {0} has different breakpoint and SNV trees'.format(patient_id)
            continue

        snv_breakpoint_nodes = pd.DataFrame(snv_breakpoint_nodes, columns=['snv_node', 'breakpoint_node'])

        snv_nodes = helpers.get_snv_nodes(patient_id)
        snv_origin_counts = snv_nodes.groupby('node')['ml_origin'].sum().reset_index().rename(columns={'ml_origin':'count'})
        snv_loss_counts = snv_nodes.groupby('node')['ml_loss'].sum().reset_index().rename(columns={'ml_loss':'count'})

        snv_counts = pd.merge(snv_origin_counts, snv_loss_counts, on='node', suffixes=('_origin', '_loss'))

        breakpoint_nodes = helpers.get_breakpoint_nodes(patient_id)
        breakpoint_origin_counts = breakpoint_nodes.groupby('node')['ml_origin'].sum().reset_index().rename(columns={'ml_origin':'count'})
        breakpoint_loss_counts = breakpoint_nodes.groupby('node')['ml_loss'].sum().reset_index().rename(columns={'ml_loss':'count'})

        breakpoint_counts = pd.merge(breakpoint_origin_counts, breakpoint_loss_counts, on='node', suffixes=('_origin', '_loss'))

        event_counts = snv_breakpoint_nodes.copy()

        event_counts = event_counts.merge(snv_counts, left_on='snv_node', right_on='node')\
                                   .drop(['node'], axis=1)\
                                   .rename(columns={'count_origin':'count_origin_snv',
                                                    'count_loss':'count_loss_snv'})

        event_counts = event_counts.merge(breakpoint_counts, left_on='breakpoint_node', right_on='node')\
                                   .drop(['node'], axis=1)\
                                   .rename(columns={'count_origin':'count_origin_breakpoint',
                                                    'count_loss':'count_loss_breakpoint'})

        event_counts['patient_id'] = patient_id

        event_table.append(event_counts)

    event_table = pd.concat(event_table, ignore_index=True)

    return event_table

    
class gaussian_kde_bw(scipy.stats.gaussian_kde):
    def __init__(self, dataset, bw):
        self.covariance = bw**2.
        scipy.stats.gaussian_kde.__init__(self, dataset)
    def _compute_covariance(self):
        self.inv_cov = 1.0 / self.covariance
        self._norm_factor = np.sqrt(2*np.pi*self.covariance) * self.n


def snvs_loss_density_table(bw=1e8):

    density_table = list()

    for patient_id in helpers.get_wgss_patient_ids():

        nodes = helpers.get_snv_nodes(patient_id)

        lost = nodes[nodes['ml_loss'] == 1]
        lost = lost.drop_duplicates(['chrom', 'coord', 'ref', 'alt'])

        lost.set_index('chrom', inplace=True)
        lost['chromosome_start'] = plots.utils.chromosome_start
        lost.reset_index(inplace=True)

        lost['coord'] = lost['coord'] + lost['chromosome_start']

        density = gaussian_kde_bw(lost['coord'].values, bw)

        density_eval = lost['coord'].values
        density_eval = np.linspace(0, 3e9, 1000000)

        mean_density = density(density_eval).mean()
        count = len(lost['coord'].values)

        density_table.append((patient_id, mean_density, count))

    density_table = pd.DataFrame(density_table, columns=('patient_id', 'density', 'count'))

    return density_table


def calculate_presence_absence_genotypes(patient_id, absence_threshold=0.5):
    """ Nodes table for pyclone clusters based on presence absence.

    Create a table of maximum likelihood indicators for origin presence and
    loss at each node in the wgss tree.  Events are pyclone clusters and 
    presence at leaves is determined based on the cluster posteriors for 0
    and 1 prevalence.
    
    """

    wgss_site_ids = helpers.get_wgss_patient_tumour_site_names(patient_id)

    tree = helpers.get_snv_tree(patient_id)

    pyclone = pd.read_csv(
        helpers.results_directory + '/clonal_phylogeny/cluster_posteriors/patient_{0}.tsv.gz'.format(patient_id),
        sep='\t', compression='gzip'
    )

    pyclone.rename(columns={'cluster_id':'cluster_id', 'sample_id':'site_id'}, inplace=True)
    pyclone = pyclone[pyclone['site_id'].isin(wgss_site_ids)]
    pyclone.drop('size', axis=1, inplace=True)
    pyclone.set_index(['cluster_id', 'site_id'], inplace=True)

    def compute_posterior_probability(log_pdf, a, b):
        step_size = log_pdf.index[1] - log_pdf.index[0]
        log_step_size = pd.np.log(step_size)
        x = log_pdf.index[(log_pdf.index >= a) & (log_pdf.index <= b)]
        log_p = scipy.misc.logsumexp(log_step_size + log_pdf.loc[x])
        return log_p

    pyclone.columns = pyclone.columns.astype(float)
    
    ll_absent = pyclone.apply(compute_posterior_probability, axis=1, args=(0, absence_threshold))
    ll_present = pyclone.apply(compute_posterior_probability, axis=1, args=(absence_threshold, 1 ))
    
    ll = pd.concat([ll_absent, ll_present], axis=1)
    ll.rename(columns={0 : 'log_likelihood_absent', 1 : 'log_likelihood_present'}, inplace=True)
    
    ll = ll.T.stack().T

    tree.add_parent()
    tree.add_ancestors()

    ml_data = dollo.run.compute_max_likelihood_indicators(ll, tree, 0.5)

    nodes_table = pd.DataFrame(ml_data)
    nodes_table.reset_index(inplace=True)

    return nodes_table





import gzip
import pickle
import copy
import string
import numpy as np
import pandas as pd

import wgs_analysis.helpers as helpers
import wgs_analysis.plots as plots
import wgs_analysis.plots.trees
import wgs_analysis.algorithms as algorithms
import wgs_analysis.algorithms.tree


def load_wgss_node_prevalence(patient_id):

    node_prevalence = pd.read_csv(
        helpers.results_directory + '/clonal_phylogeny/clonal_prevalence_mcmc/wgss/clonal_prevalence/patient_{0}.tsv.gz'.format(patient_id),
        sep='\t', compression='gzip'
    )

    node_prevalence = node_prevalence.set_index(['clone_id', 'site_id'])['mean'].unstack()

    return node_prevalence


def load_wgss_genotypes(patient_id):

    genotypes = pd.read_csv(
        helpers.results_directory + '/clonal_phylogeny/clonal_prevalence_mcmc/wgss/nodes/patient_{0}.tsv.gz'.format(patient_id),
        sep='\t', compression='gzip'
    )

    genotypes.rename(columns={'event_id':'cluster_id'}, inplace=True)

    return genotypes


def load_wgss_tree(patient_id):

    tree_filename = helpers.results_directory + '/clonal_phylogeny/clonal_prevalence_mcmc/wgss/tree/patient_{0}.pickle.gz'.format(patient_id)

    with gzip.open(tree_filename, 'r') as f:
        tree = pickle.load(f)

    return tree


def load_sc_node_prevalence(patient_id):

    node_prevalence = pd.read_csv(
        helpers.results_directory + '/clonal_phylogeny/clonal_prevalence_mcmc/single_cell/clonal_prevalence/patient_{0}.tsv.gz'.format(patient_id),
        sep='\t', compression='gzip'
    )

    node_prevalence = node_prevalence.set_index(['clone_id', 'site_id'])['mean'].unstack()

    return node_prevalence


def load_sc_genotypes(patient_id):

    genotypes = pd.read_csv(
        helpers.results_directory + '/pyclone_single_cell_dollo/patient_{0}/nodes.tsv.gz'.format(patient_id),
        sep='\t', compression='gzip',
    )

    genotypes.rename(columns={'event_id':'cluster_id'}, inplace=True)

    return genotypes


def load_sc_tree(patient_id):

    tree = pickle.load(gzip.open(
        helpers.results_directory + '/clonal_phylogeny/clonal_prevalence_mcmc/single_cell/tree/patient_{0}.pickle.gz'.format(patient_id)))
    
    return tree


def get_wgss_pyclone_cluster_id(patient_id):
    data = helpers.read_table_data(patient_id, 'clonal_analysis/pyclone_wgss')
    data.set_index(['chrom', 'coord', 'ref', 'alt'], inplace=True)
    data = data.idxmax(axis=1).astype(int)
    data.name = 'cluster_id'
    data = data.reset_index()
    return data


def load_clone_data(patient_id):
    """ Load clone data for a patient.

    Args:
        patient_id (str): patient identifier

    Returns:
        dollo.trees.TreeNode: clone tree
        pandas.DataFrame: genotype table
        pandas.DataFrame: node prevalence table

    Load single cell data if available, otherwise wgss based clones.

    The leaves of the clone tree are the full set of clones.  All internal nodes
    have a descendent leaf with no mutations on the branch to that leaf.

    The genotype table will have node labels as row indices and event ids (ids of 
    snv/breakpoint clusters) as column labels.  Each 0/1 entry represents the
    presence or absence of the event for that node.  

    The node prevalence table table should have sample ids as row indices and leaf
    names as column labels.

    """

    # Load data
    if int(patient_id) in helpers.single_cell_patients:
        genotypes = load_sc_genotypes(patient_id)
        tree = load_sc_tree(patient_id)
        node_prevalence = load_sc_node_prevalence(patient_id)

    else:
        genotypes = load_wgss_genotypes(patient_id)
        tree = load_wgss_tree(patient_id)
        node_prevalence = load_wgss_node_prevalence(patient_id)

    # Filter normal clone, excluded samples
    site_names = helpers.get_patient_tumour_site_names(patient_id)
    node_prevalence = node_prevalence.loc[node_prevalence.index.astype(str) != 'normal', site_names]
    node_prevalence.index = node_prevalence.index.astype(int)

    # Patient 2 FFPE filtered
    if int(patient_id) == 2:
        wgs_site_names = helpers.get_wgss_patient_tumour_site_names(patient_id)
        node_prevalence = node_prevalence[wgs_site_names]

    # Renormalize
    node_prevalence = node_prevalence / node_prevalence.sum(axis=0)

    # Calculate tree cluster ids
    cluster_ids = set()
    for node in tree.nodes:
        cluster_ids.update(node.genotype)

    # Filter nodes table based on cluster ids
    genotypes = genotypes[genotypes['cluster_id'].isin(cluster_ids)]

    # Remove branches with no mutation
    tree = algorithms.tree.simplify_tree_by_genotype(tree, genotypes)

    # Add leaves representing internal nodes to the tree
    tree = plots.trees.add_internal_node_leaves(tree)

    # Letter names for leaves
    for leaf, name in zip(reversed(list(tree.leaves)), string.ascii_uppercase):
        leaf.name = name

    # Node prevalence table with columns as leaf names, row indices as site ids
    leaf_name = dict([(leaf.label, leaf.name) for leaf in tree.leaves])

    node_prevalence['leaf'] = pd.Series(leaf_name).reindex(node_prevalence.index)

    node_prevalence = (
        node_prevalence
        .set_index('leaf')
        .T
        .rename(index=lambda site_name: helpers.get_plot_id(patient_id, site_name))
    )

    # Genotype table, node labels as row indices, cluster ids as column names
    genotypes = genotypes.set_index(['node', 'cluster_id'])['ml_presence'].unstack()

    return tree, genotypes, node_prevalence


def calculate_cluster_origin(tree, genotypes):
    """ Calculate the origin node of each pyclone cluster.

    Args:
        tree (dollo.TreeNode): tree to simplify
        genotypes (pandas.DataFrame): presence absence table

    Returns:
        dict: clone (leaf name) keyed by cluster

    The presence absence table should have node labels as row indices and pyclone
    cluster id as column name.  Values are 0/1 for presence/absence.

    """

    cluster_presence = (
        genotypes
        .stack()
        .reset_index()
        .rename(columns={0:'presence'})
    )

    cluster_presence = (
        cluster_presence[cluster_presence['presence'] == 1]
        .drop(['presence'], axis=1)
    )

    for node in tree.nodes:
        node.clusters = set()

    for node_label, cluster_id in cluster_presence.values:
        for node in tree.nodes:
            if node.label == node_label:
                node.clusters.add(cluster_id)

    tree.add_parent()

    # Calculate the number of new SNVs that
    # distinguish parent from child
    for node in tree.nodes:
        node.origins = set()
        for cluster_id in node.clusters:
            if node.parent is None or cluster_id not in node.parent.clusters:
                node.origins.add(cluster_id)

    # Propogate to leaves representing internal nodes
    for node in tree.nodes:
        if node.parent is not None and node.parent.label == node.label:
            node.origins = node.parent.origins

    # Create a map between cluster id and leaf name
    cluster_origin = dict()
    for leaf in tree.leaves:
        for cluster_id in leaf.origins:
            cluster_origin[cluster_id] = leaf.name

    return cluster_origin


def calculate_num_gained(tree, genotypes, sizes):
    """ Count number of gained SNVs per clone.

    Args:
        tree (dollo.TreeNode): tree to simplify
        genotypes (pandas.DataFrame): presence absence table
        sizes (mapping): cluster sizes

    Returns:
        dict: number of gained SNVs keyed by clone (leaf name)

    The presence absence table should have node labels as row indices and pyclone
    cluster id as column name.  Values are 0/1 for presence/absence.

    The sizes should be a mapping type with pyclone cluster ids as keys and cluster
    sizes as values.

    > Does not count losses

    """

    cluster_origin = calculate_cluster_origin(tree, genotypes)

    num_gained = dict([(leaf.name, 0) for leaf in tree.leaves])
    for cluster_id, size in sizes.iteritems():
        if cluster_id in cluster_origin:
            num_gained[cluster_origin[cluster_id]] += size

    return num_gained


def load_prevalence_data(patient_id, clade=False):
    """ Load prevalences for a patient.

    Args:
        patient_id (int): patient

    KwArgs:
        clade (bool): clade prevalences

    Returns:
        pandas.DataFrame clone prevalence dataframe.

    Return dataframe has columns clone_id, site_id, num_gained_wgss, num_gained_deepseq, prevalence.

    """

    pyclone = pd.read_csv(
        helpers.results_directory + '/clonal_phylogeny/cluster_posteriors/patient_{0}.tsv.gz'.format(patient_id),
        sep='\t', compression='gzip'
    )
    
    pyclone_sizes = (
        pyclone
        .drop_duplicates(['cluster_id', 'size'])
        .set_index(['cluster_id'])['size']
    )

    wgss_sizes = helpers.get_pyclone_wgss_cluster_sizes(patient_id)

    tree, genotypes, node_prevalence = load_clone_data(patient_id)

    if clade:
        tree.add_parent()

        clade_prevalence = node_prevalence.copy() * 0
        node_clone = tree.get_attr_map('name')
        for node in tree.nodes:
            if node.parent is not None and node.parent.label == node.label:
                continue
            for leaf in node.leaves:
                clade_prevalence[node_clone[node.label]] += node_prevalence[leaf.name]

        prevalence = clade_prevalence
        
    else:
        prevalence = node_prevalence

    num_gained_wgss = calculate_num_gained(tree, genotypes, wgss_sizes)
    num_gained_deepseq = calculate_num_gained(tree, genotypes, pyclone_sizes)

    tissue_source = helpers.sample_info.loc[
        (helpers.sample_info['patient_id'] == int(patient_id)) &
        (helpers.sample_info['malignant'] == 'yes'),
        ['plot_id', 'tissue_source']
    ].rename(columns={'plot_id':'site_id'})

    anatomy_id = helpers.sample_info.loc[
        (helpers.sample_info['patient_id'] == int(patient_id)) &
        (helpers.sample_info['malignant'] == 'yes'),
        ['plot_id', 'anatomy_id']
    ].rename(columns={'plot_id':'site_id'})

    prevalence_table = (
        prevalence
        .stack()
        .reset_index()
        .rename(columns={0:'prevalence', 'leaf':'clone_id'})
        .merge(pd.DataFrame({'num_gained_wgss':num_gained_wgss}), left_on='clone_id', right_index=True)
        .merge(pd.DataFrame({'num_gained_deepseq':num_gained_deepseq}), left_on='clone_id', right_index=True)
        .merge(tissue_source, on='site_id', how='outer')
        .merge(anatomy_id, on='site_id', how='outer')
        .dropna()
    )

    return prevalence_table


def calculate_sample_classification(tree, node_prevalence, presence_threshold=0.01):
    """ Calculate branching classification for each sample.

    Args:
        tree (dollo.TreeNode): clone tree
        node_prevalence (pandas.DataFrame): node prevalence table

    KwArgs:
        presence_threshold (float): prevalence threshold for calling presence

    Returns:
        pandas.Series: classification with sample index

    Classify the branching relationship of clones present in each sample.

    """

    tree = copy.deepcopy(tree)
    clone_node = dict([(leaf.name, leaf.label) for leaf in tree.leaves])

    # Remove leaves representing internal nodes
    for node in list(tree.nodes):
        children = list()
        for child in node.children:
            if child.label != node.label:
                children.append(child)
        node.children = children

    def classify_sample_tree(prv):
        site_clones = prv[prv > presence_threshold].index.values
        site_labels = [clone_node[a] for a in site_clones]
        site_tree = algorithms.tree.simplify_tree(copy.deepcopy(tree), site_labels)
        return algorithms.tree.classify_tree(site_tree)

    sample_classification = node_prevalence.apply(classify_sample_tree, axis=1)

    return sample_classification



def load_sample_composition_data(patient_id, presence_threshold=0.01):
    """ Calculate a table of clonal composition statistics per sample.

    Args:
        patient_id (str): patient
        presence_threshold (float): minimum prevalence for clonal presence

    Returns:
        pandas.DataFrame: table of per sample info with columns
          'site_id', 'divergence', 'entropy', 'sample_class'

    'sample_class' is one of 'chain', 'branched', 'pure'

    """

    pyclone = pd.read_csv(
        helpers.results_directory + '/clonal_phylogeny/cluster_posteriors/patient_{0}.tsv.gz'.format(patient_id),
        sep='\t', compression='gzip'
    )
    
    pyclone_sizes = (
        pyclone
        .drop_duplicates(['cluster_id', 'size'])
        .set_index(['cluster_id'])['size']
    )

    wgss_sizes = helpers.get_pyclone_wgss_cluster_sizes(patient_id)

    tree, genotypes, node_prevalence = load_clone_data(patient_id)

    data = node_prevalence.stack().reset_index().rename(columns={0:'prevalence', 'leaf':'clone'})
    data = data[data['prevalence'] > presence_threshold]

    label_clone = tree.get_attr_map('name')

    # Create table of 'cluster_id', 'present', 'size' indexed by clone
    clone_genotypes = genotypes.reindex(label_clone.keys()).rename(index=label_clone)
    clone_genotypes.index.name = 'clone'
    clone_genotypes = clone_genotypes.stack().reset_index().rename(columns={0:'present'})
    clone_genotypes = clone_genotypes[clone_genotypes['present'] > 0]
    clone_genotypes = clone_genotypes.set_index('clone').sort_index()

    site_stats = list()

    for site_id, site_data in data.groupby('site_id'):
        clones = set(site_data['clone'].values)
        union = set()
        intersection = set(clone_genotypes['cluster_id'].unique())
        for clone in clones:
            clusters = clone_genotypes.loc[clone:clone, 'cluster_id'].values
            union = union.union(clusters)
            intersection = intersection.intersection(clusters)
        union_size = sum([wgss_sizes[a] for a in union])
        intersection_size = sum([wgss_sizes[a] for a in intersection])
        divergence = 1.0 - float(intersection_size) / float(union_size)
        entropy = -np.sum(site_data['prevalence'] * np.log(site_data['prevalence']))

        site_stats.append((site_id, divergence, entropy))

    sample_data = pd.DataFrame(site_stats, columns=['site_id', 'divergence', 'entropy'])

    sample_data.set_index('site_id', inplace=True)
    sample_data['sample_class'] = calculate_sample_classification(tree, node_prevalence)
    sample_data.reset_index(inplace=True)

    return sample_data





import argparse
import os
import socket
import sys
import subprocess
import time

import design_utils


class BlatServer(object):
    ''' 
    Run a gfServer in a context manager.
    Block until the server is up and accepting connections from clients.
    '''
    def __init__(self, genome_filename):
        self.genome_filename = os.path.abspath(genome_filename)
    def __enter__(self):
        self.gfserver_proc = None
        server_exists = False
        try:
            socket.socket().connect(('localhost', 8899))
            server_exists = True
        except socket.error:
            pass
        if server_exists:
            print 'A gfServer process is already running'
            return
        print 'Running gfServer for genome {}'.format(self.genome_filename)
        print 'To start gfServer directly, run:'
        print 'python {} {}'.format(os.path.realpath(__file__), os.path.realpath(self.genome_filename))
        self.gfserver_proc = subprocess.Popen(['gfServer', 'start', 'localhost', '8899', '-stepSize=5', self.genome_filename + '.2bit'], stdout=subprocess.PIPE)
        with design_utils.TempDirectory() as temps_dir:
            temp_fasta_filename = os.path.join(temps_dir, 'sequence.fa')
            temp_psl_filename = os.path.join(temps_dir, 'align.psl')
            with open(temp_fasta_filename, 'w') as temp_fasta_file:
                temp_fasta_file.write('>1\nCAGGCTTTTAAATTGGCTTTGATGG\n')
            while True:
                try:
                    subprocess.check_call(['gfClient', 'localhost', '8899', '/', temp_fasta_filename, temp_psl_filename], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    break
                except subprocess.CalledProcessError:
                    pass
                sys.stderr.write('Waiting 10s for gfServer to start\n')
                time.sleep(10)
    def __exit__(self ,type, value, traceback):
        if self.gfserver_proc is not None:
            self.gfserver_proc.kill()


if __name__ == '__main__':
    argparser = argparse.ArgumentParser()
    argparser.add_argument('genome', help='genome fasta filename')
    args = argparser.parse_args()
    with BlatServer(args.genome):
        while True:
            time.sleep(100)





import seaborn as sb
import numpy as np


def setup_plot():
    """
    Set up matplotlib in a consistent way between plots.
    """
    sb.set_style('ticks', {'font.sans-serif':['Helvetica']})


def setup_axes(ax):
    """
    Set up axes in a consistent way between plots.  Spines are at left and bottom, offset by 10pts.
    Other spines not visible.  Ticks only for left and bottom.
    """
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    
    ax.xaxis.tick_bottom()
    ax.yaxis.tick_left()
    
    ax.xaxis.grid(True, which="major", linestyle=':')
    ax.yaxis.grid(True, which="major", linestyle=':')

    return ax


def shift_lims(ax, x, y):
    """
    Shift the x and y limits in pixel space to make room for plot elements on the boundary of the plot.
    """
    x_dev, y_dev = ax.transData.inverted().transform_point((x,y)) - ax.transData.inverted().transform_point((0,0))

    ax.set_xlim([ax.get_xlim()[0] - x_dev, ax.get_xlim()[1] + x_dev])
    ax.set_ylim([ax.get_ylim()[0] - y_dev, ax.get_ylim()[1] + y_dev])


def _filter_ticks(ticks, data_min, data_max):
    ticks_before = list(filter(lambda a: a < data_min, ticks))
    ticks_within = list(filter(lambda a: a >= data_min and a <= data_max, ticks))
    ticks_after = list(filter(lambda a: a > data_max, ticks))

    ticks = []
    if len(ticks_before) > 0:
        ticks.append(max(ticks_before))
    ticks.extend(ticks_within)
    if len(ticks_after) > 0:
        ticks.append(min(ticks_after))

    return ticks


def set_xlim_filter_ticks(ax, xmin, xmax):
    """
    Set x limits and remove ticks outside of limits.
    """
    ax.set_xlim((xmin, xmax))
    ax.set_xticks(_filter_ticks(ax.get_xticks(), xmin, xmax))


def set_ylim_filter_ticks(ax, ymin, ymax):
    """
    Set y limits and remove ticks outside of limits.
    """
    ax.set_ylim((ymin, ymax))
    ax.set_yticks(_filter_ticks(ax.get_yticks(), ymin, ymax))


def trim_spines_to_ticks(ax):
    """
    Trim spines to start and end at first and last ticks.
    """
    xticks = np.concatenate((ax.get_xticks(), ax.get_xticks(minor=True)))
    ax.spines['bottom'].set_bounds(min(xticks), max(xticks))
    yticks = np.concatenate((ax.get_yticks(), ax.get_yticks(minor=True)))
    ax.spines['left'].set_bounds(min(yticks), max(yticks))





'''
CNV annotation functionality

Annotations for cnv data.

Functions take pandas tables and add additional columns.
'''

import collections
import numpy as np
import pandas as pd

import wgs_analysis.algorithms as algorithms
import wgs_analysis.algorithms.merge


def label_presence(node, presence):
    """ Add present label to nodes

    Args:
        node (wgs_analysis.trees.TreeNode): tree node
        presence (dict): dictionary of presences by leaf name

    Add the present label to nodes.  By default nodes are labeled absent.  If any 
    of a nodes children have been labeled present, label that node as present.

    """

    if node.is_leaf:

        node.present = presence[node.name]

    else:

        node.present = 0

        for child in node.children:

            label_presence(child, presence)

            if child.present:
                node.present = 1


class InconsistentError(Exception):
    pass

def check_consistency(node):
    """ Check presence consistency

    Args:
        node (wgs_analysis.trees.TreeNode): tree node

    Raises:
        InconsistentError

    Check that if a node is labeled as present, so are all its children,
    otherwise raise an exception.

    """
    for child in node.children:

        if node.present and not child.present:
            raise InconsistentError()

        check_consistency(child)


def annotate_presence_timing(variants, tree, column_name, presence_f):
    """ Annotate timing of a presence absence feature

    Args:
        cnv (pandas.DataFrame): variant_site table
        tree (wgs_analysis.trees.TreeNode): tree
        column_name (str): column name to annotate
        presence_f (callable): function to calculate a dictionary of presences by site

    Returns:
        pandas.DataFrame: original table with additional column

    Annotate a variant table with the inferred timing of a presence absence event.

    """

    def calculate_timing(df):

        df = df.set_index('site_id')

        is_present = presence_f(df).to_dict()

        label_presence(tree, is_present)

        try:
            check_consistency(tree)
        except InconsistentError:
            return 'inconsistent'

        if np.sum(is_present) == len(df.index):
            return 'ancestral'
        elif np.sum(is_present) == 1:
            return 'private'
        else:
            return 'descendent'

    variants.set_index(['chrom', 'coord'], inplace=True)
    variants[column_name] = variants.groupby(level=[0, 1]).apply(calculate_timing)
    variants.reset_index(inplace=True)

    return variants


def annotate_loh_timing(variants, tree):
    """ Add loh_timing column

    Args:
        cnv (pandas.DataFrame): variant_site table
        tree (wgs_analysis.trees.TreeNode): tree

    Returns:
        pandas.DataFrame: original table with `loh_timing` column

    Annotate a variant table with the inferred timing of LOH events.

    """

    def calculate_loh_presence(df):
        return (df.minor == 0) * 1

    return annotate_presence_timing(variants, tree, 'loh_timing', calculate_loh_presence)


def annotate_hdel_timing(variants, tree):
    """ Add hdel_timing column

    Args:
        cnv (pandas.DataFrame): variant_site table
        tree (wgs_analysis.trees.TreeNode): tree

    Returns:
        pandas.DataFrame: original table with `hdel_timing` column

    Annotate a variant table with the inferred timing of homozygous deletion events.

    """

    def calculate_loh_presence(df):
        return (df.minor == 0) * 1

    return annotate_presence_timing(variants, tree, 'hdel_timing', calculate_loh_presence)


def annotate_cnv_type(copies):
    """ Add various cnv type columns

    Args:
        copies (pandas.DataFrame): cnv table

    Returns:
        pandas.DataFrame: original table with additional column
    """

    # LOH status for each segment
    copies['loh'] = (copies['minor'] < 0.5) * 1

    # Homozygous deletion status for each segment
    copies['hdel'] = (copies['major'] < 0.5) * 1

    # Subclonal LOH status (only if not loh)
    copies['subclonal_loh'] = (np.minimum(copies['minor'] + copies['minor_sub'], copies['major'] + copies['major_sub']) < 0.5) * 1
    copies['subclonal_loh'] = copies['subclonal_loh'] * (1 - copies['loh'])

    # Subclonal homozygous deletion status (only if not dominant hdel)
    copies['subclonal_hdel'] = (np.maximum(copies['minor'] + copies['minor_sub'], copies['major'] + copies['major_sub']) < 0.5) * 1
    copies['subclonal_hdel'] = copies['subclonal_hdel'] * (1 - copies['hdel'])

    return copies


def annotate_position_cnv(positional, cnv):
    """ Add cnv columns to positional data

    Args:
        positional (pandas.DataFrame): positions table with 'site_id', 'chrom', 'coord' columns
        cnv (pandas.DataFrame): cnv data with 'site_id', 'chrom', 'start', 'end' and additional copies columns

    Returns:
        pandas.DataFrame: positions table with additional columns from cnv table
    """
    merged = algorithms.merge.position_segment_merge(positional, cnv)

    cnv = cnv.merge(merged, on=['chrom', 'start', 'end']).drop(['start', 'end'], axis=1)

    positional = positional.merge(cnv, on=['chrom', 'coord'], how='left')

    return positional










'''
# Gene matrix table generation

Functions for generating tables of mutations/rearrangements for use with gene matrix plots and selection of interesting
candidates.

'''

import pandas as pd
import numpy as np

import wgs_analysis.helpers as helpers
import wgs_analysis.tables as tables


def sort_matrix(matrix):
    matrix['sort_value'] = matrix.apply(lambda x: get_sort_value(x), axis=1)
    
    matrix = matrix.sort(columns='sort_value', ascending=False)
    
    matrix = matrix.drop('sort_value', axis=1)
    
    return matrix


def get_sort_value(row):
    binary_str = ''
    
    for x in row:
        if np.isnan(x):
            x = 1
        
        binary_str += str(int(x))
    
    return int(binary_str, 2)


def _build_gene_matrix(data):

    gene_matrix = data.set_index(['event_name', 'gene_name', 'site_id'])['present'].unstack(level=-1)

    gene_matrix = gene_matrix.reset_index().drop('event_name', axis=1)

    gene_matrix = gene_matrix.drop_duplicates()

    gene_matrix = gene_matrix.set_index('gene_name')

    gene_matrix = sort_matrix(gene_matrix)
    
    gene_matrix = sort_matrix(gene_matrix.T).T
    
    gene_matrix = gene_matrix.rename(columns=lambda a: helpers.plot_ids[a])
    
    return gene_matrix


def _is_snv_present(row, min_depth):
    depth = row['ref_counts'] + row['alt_counts']
    
    if depth < min_depth:
        return np.nan
    
    elif row['alt_counts'] > 0:
        return 1
    
    else:
        return 0


def _is_indel_present(row):
    if row['alt_counts'] > 0:
        return 1
    
    else:
        return 0


def _is_breakpoint_present(row):
    if np.isnan(row['num_span']):
        return np.nan
    
    elif row['num_span'] > 0:
        return 1
    
    else:
        return 0
    

def load_filtered_snvs(patient_id):
    data = helpers.get_snvs(patient_id)

    data = data[data['site_id'].isin(helpers.get_patient_tumour_site_names(patient_id))]

    data = data[data.snpeff_impact.isin(['HIGH', 'MODERATE'])]

    data = data[data.gene_name.isin(helpers.target_snv_genes)]

    return data


def load_snv_matrix(patient_id, min_present_depth=10):
    data = load_filtered_snvs(patient_id)

    if len(data.index) == 0:
        return pd.DataFrame()

    def event_name(row):
        return ':'.join(['snv', str(row['chrom']), str(row['coord'])])
    
    data['event_name'] = data.apply(event_name, axis=1)
    
    data['present'] = data.apply(lambda x : _is_snv_present(x, min_present_depth), axis=1)
  
    data = data[['site_id', 'event_name', 'gene_name', 'present']].drop_duplicates()

    if len(data.index) == 0:
        return pd.DataFrame()

    return _build_gene_matrix(data)


def load_filtered_breakpoints(patient_id):
    data = helpers.get_breakpoints(patient_id)

    expression = helpers.get_recentred_expression(patient_id)

    data = tables.rearrangement.expression_correlated(data, expression)

    data = data[data.gene_name.isin(helpers.target_genes)]

    return data


def load_breakpoint_matrix(patient_id):
    data = load_filtered_breakpoints(patient_id)

    def event_name(row):
        return ':'.join(['brk', str(row['cluster_id']), str(row['cluster_end'])])
    
    data['event_name'] = data.apply(event_name, axis=1)
    
    data['present'] = data.apply(_is_breakpoint_present, axis=1) 

    data = data[['site_id', 'event_name', 'gene_name', 'present']].drop_duplicates()
    
    if len(data.index) == 0:
        return pd.DataFrame()    
    
    return _build_gene_matrix(data)


def load_filtered_indels(patient_id):
    data = helpers.get_indels(patient_id)

    data = data[data['site_id'].isin(helpers.get_patient_tumour_site_names(patient_id))]
    
    data = data[data.gene_name.isin(helpers.target_genes)]

    data.set_index(['chrom', 'coord', 'ref', 'alt'], inplace=True)
    data['max_alt_counts'] = data.set_index('site_id', append=True)['alt_counts'].unstack().max(axis=1)
    data.reset_index(inplace=True)

    data = data[data['max_alt_counts'] > 0]

    return data


def load_indel_matrix(patient_id, min_present_depth=10):
    data = load_filtered_indels(patient_id)

    if len(data.index) == 0:
        return pd.DataFrame()    

    index_columns = ['chrom', 'coord', 'ref', 'alt', 'gene_name']

    data = data.set_index(index_columns + ['site_id'])[['ref_counts', 'alt_counts']].unstack().fillna(0).stack().reset_index()

    def event_name(row):
        return ':'.join(['indel', str(row['chrom']), str(row['coord'])])
    
    data['event_name'] = data.apply(event_name, axis=1)
    
    data['present'] = data.apply(_is_indel_present, axis=1)

    data = data[['site_id', 'event_name', 'gene_name', 'present']].drop_duplicates()
    
    if len(data.index) == 0:
        return pd.DataFrame()    
    
    return _build_gene_matrix(data)


def load_cn_matrix(patient_id, cnv_type):

    if cnv_type not in ('amp', 'hdel'):
        raise ValueError(cnv_type)

    gene_regions = helpers.load_gene_regions()

    if cnv_type == 'amp':
        target_genes = helpers.target_amp_genes
    elif cnv_type == 'hdel':
        target_genes = helpers.target_del_genes

    target_gene_regions = gene_regions[gene_regions['gene_name'].isin(target_genes)]

    copies = helpers.get_copies(patient_id)

    stats = helpers.get_copies_stats(patient_id)

    site_ids = list(stats['site_id'].unique())

    stats['ploidy'] = (stats['minor_size'] + stats['major_size']) / 3e9

    copies = copies.merge(stats[['site_id', 'ploidy']], left_on='site_id', right_on='site_id')

    copies = copies.rename(columns={'site_id':'site_id'})

    data = list()

    for _, row in target_gene_regions.iterrows():

        ovlap = copies[(copies['chrom'] == row['chrom']) &
                       (copies['end'] >= row['start']) &
                       (copies['start'] <= row['end'])]

        if len(ovlap.index) == 0:
            continue

        if cnv_type == 'amp':
            variant = ovlap.loc[ovlap['major'] > 3. * ovlap['ploidy'], ['site_id']].drop_duplicates()
        elif cnv_type == 'hdel':
            variant = ovlap.loc[ovlap['major'] == 0, ['site_id']].drop_duplicates()

        if len(variant.index) == 0:
            continue

        variant['gene_name'] = row['gene_name']

        data.append(variant)

    if len(data) == 0:
        return pd.DataFrame()

    data = pd.concat(data, ignore_index=True)

    data['present'] = 1

    data = data.set_index(['gene_name', 'site_id'])['present']\
               .unstack()\
               .reindex(columns=site_ids)\
               .fillna(0)\
               .stack()\
               .reset_index()
    data.columns = ['gene_name', 'site_id', 'present']

    data['event_name'] = data['gene_name'] + '_' + cnv_type

    if len(data.index) == 0:
        return pd.DataFrame()    
    
    return _build_gene_matrix(data)

#=======================================================================================================================
# Deepseq
#=======================================================================================================================
def load_filtered_deeseq_snvs(patient_id):
    data = helpers.get_deepseq_snvs(patient_id)

    data = data[data['site_id'].isin(helpers.get_patient_tumour_site_names(patient_id))]

    data = data[data.snpeff_impact.isin(['HIGH', 'MODERATE'])]
    
    data = data[~data.primer_id.str.contains('amplicrazy')]

    return data

def load_deepseq_snv_matrix(patient_id, p_value_threshold=1e-6):
    data = load_filtered_deeseq_snvs(patient_id)

    if len(data.index) == 0:
        return pd.DataFrame()

    def event_name(row):
        return ':'.join(['snv', str(row['chrom']), str(row['coord'])])
    
    data['event_name'] = data.apply(event_name, axis=1)
    
    data['present'] = data.apply(lambda x: _is_deeqseq_snv_present(x, p_value_threshold), axis=1)
  
    data = data[['site_id', 'event_name', 'gene_name', 'present']].drop_duplicates()

    if len(data.index) == 0:
        return pd.DataFrame()

    return sort_matrix(_build_gene_matrix(data))

def _is_deeqseq_snv_present(row, threshold):
    return row['alt_p_value'] < threshold

def load_filtered_deeqpseq_breakpoints(patient_id, min_counts=5):
    data = helpers.get_deepseq_breakpoints(patient_id)

    germline = data.loc[
        (data['site_id'] == 'normal_blood') &
        (data['count'] >= min_counts),
        'seq_id'].unique()

    data = data[~data['seq_id'].isin(germline)]
    
    data = data[data['site_id'].isin(helpers.get_patient_tumour_site_names(patient_id))]
    
    return data

def load_deepseq_breakpoint_matrix(patient_id, min_counts=5):
    data = load_filtered_deeqpseq_breakpoints(patient_id, min_counts=min_counts)

    if len(data.index) == 0:
        return pd.DataFrame()

    def event_name(row):
        return row['seq_id']

    data['event_name'] = data.apply(event_name, axis=1)

    data['gene_name'] = (
        data['gene_name'] + ' ' +
        data['chrom_1'] + data['strand_1'] + data['coord_1'].astype(str) + ':' + 
        data['chrom_2'] + data['strand_2'] + data['coord_2'].astype(str))

    data['present'] = data.apply(lambda x: _is_deepseq_breakpoint_present(x, min_counts), axis=1)
  
    data = data[['site_id', 'event_name', 'gene_name', 'present']].drop_duplicates()

    if len(data.index) == 0:
        return pd.DataFrame()

    return sort_matrix(_build_gene_matrix(data))

def _is_deepseq_breakpoint_present(row, threshold):
    return row['count'] >= threshold


'''
# SNV plotting

'''

import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import seaborn

import wgs_analysis.plots as plots
import wgs_analysis.plots.colors
import wgs_analysis.refgenome as refgenome


def snv_adjacent_density_plot(ax, snvs, color_by_chromosome=True, alpha=0.5, s=5, lw=0, **kwargs):

    snvs = snvs.drop_duplicates(['chrom', 'coord'])

    snvs = snvs.loc[(snvs['chrom'].isin(refgenome.info.chromosomes))]

    snvs.set_index('chrom', inplace=True)
    snvs['chromosome_start'] = refgenome.info.chromosome_start
    snvs['chromosome_color'] = pd.Series(plots.colors.create_chromosome_color_map())
    snvs.reset_index(inplace=True)

    snvs['plot_coord'] = snvs['coord'] + snvs['chromosome_start']

    assert not snvs['chromosome_color'].isnull().any()

    facecolors = None
    edgecolors = None
    if color_by_chromosome:
        facecolors = list(snvs['chromosome_color'])
        edgecolors = list(snvs['chromosome_color'])

    ax.scatter(snvs['plot_coord'], snvs['adjacent_density'], 
                facecolors=facecolors,
                edgecolors=edgecolors,
                alpha=alpha, s=s, lw=lw, **kwargs)

    ax.set_xlim(min(snvs['plot_coord']), max(snvs['plot_coord']))
    ax.set_xticks(refgenome.info.chromosome_mid, minor=False)
    ax.set_xticklabels(refgenome.info.chromosomes, minor=False)
    ax.set_xticks(refgenome.info.chromosome_end, minor=True)
    ax.set_xticklabels([], minor=True)
    ax.grid(False, which='major')

    ax.set_ylim(-0.0001, 1.1 * snvs['adjacent_density'].fillna(0.).max())
    ax.set_ylabel('snv density')


def snv_adjacent_distance_plot(ax, snvs, color_by_chromosome=True, alpha=0.5, s=5, lw=0, **kwargs):

    snvs = snvs.drop_duplicates(['chrom', 'coord'])

    snvs = snvs.loc[(snvs['chrom'].isin(refgenome.info.chromosomes))]

    snvs.set_index('chrom', inplace=True)
    snvs['chromosome_start'] = refgenome.info.chromosome_start
    snvs['chromosome_color'] = pd.Series(plots.colors.create_chromosome_color_map())
    snvs.reset_index(inplace=True)

    snvs['plot_coord'] = snvs['coord'] + snvs['chromosome_start']

    assert not snvs['chromosome_color'].isnull().any()

    snvs['adjacent_distance_log'] = snvs['adjacent_distance'].apply(np.log10)

    facecolors = None
    edgecolors = None
    if color_by_chromosome:
        facecolors = list(snvs['chromosome_color'])
        edgecolors = list(snvs['chromosome_color'])

    ax.scatter(
        snvs['plot_coord'],
        snvs['adjacent_distance_log'], 
        facecolors=facecolors,
        edgecolors=edgecolors,
        alpha=alpha, s=s, lw=lw, **kwargs)

    ax.set_xlabel('chromosome')
    ax.set_xlim(min(snvs['plot_coord']), max(snvs['plot_coord']))
    ax.set_xticks([0] + list(wgs_analysis.refgenome.info.chromosome_end.values))
    ax.set_xticklabels([])
    ax.xaxis.tick_bottom()
    ax.yaxis.tick_left()
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(wgs_analysis.refgenome.info.chromosome_mid))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(wgs_analysis.refgenome.info.chromosomes))

    ax.set_ylim(-0.0001, 1.1 * snvs['adjacent_distance_log'].fillna(0.).max())
    ax.set_ylabel('Distance between mutations (log10)')

    seaborn.despine(trim=True)


def snv_genome_vaf_plot(ax, snvs):
    snvs = snvs.drop_duplicates(['chrom', 'coord'])

    snvs = snvs.loc[(snvs['chrom'].isin(refgenome.info.chromosomes))]

    snvs.set_index('chrom', inplace=True)
    snvs['chromosome_start'] = refgenome.info.chromosome_start
    snvs['chromosome_color'] = pd.Series(plots.colors.create_chromosome_color_map())
    snvs.reset_index(inplace=True)

    snvs['plot_coord'] = snvs['coord'] + snvs['chromosome_start']

    assert not snvs['chromosome_color'].isnull().any()

    ax.scatter(
        snvs['plot_coord'],
        snvs['vaf'], 
        facecolors=list(snvs['chromosome_color']),
        edgecolors=list(snvs['chromosome_color']),
        alpha=0.5, s=5, lw=0)

    ax.set_xlabel('chromosome')
    ax.set_xlim(min(snvs['plot_coord']), max(snvs['plot_coord']))
    ax.set_xticks([0] + list(wgs_analysis.refgenome.info.chromosome_end.values))
    ax.set_xticklabels([])
    ax.xaxis.tick_bottom()
    ax.yaxis.tick_left()
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(wgs_analysis.refgenome.info.chromosome_mid))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(wgs_analysis.refgenome.info.chromosomes))

    ax.set_ylim(-0.01, 1.01)
    ax.set_ylabel('VAF')

    seaborn.despine(trim=True)


def snv_count_plot(ax, snvs, bin_size=10000000, full_genome=True):
    snvs = snvs.loc[(snvs['chrom'].isin(refgenome.info.chromosomes))].copy()

    snvs.set_index('chrom', inplace=True)
    snvs['chromosome_start'] = refgenome.info.chromosome_start
    snvs.reset_index(inplace=True)

    snvs['plot_position'] = snvs['coord'] + snvs['chromosome_start']
    snvs['plot_bin'] = snvs['plot_position'] / bin_size
    snvs['plot_bin'] = snvs['plot_bin'].astype(int)

    if full_genome:
        plot_bin_starts = np.arange(0, refgenome.info.chromosome_end.max(), bin_size)
    else:
        min_x, max_x = snvs['plot_position'].min(), snvs['plot_position'].max()
        plot_bin_starts = np.arange(min_x, max_x, bin_size)
    plot_bins = (plot_bin_starts / bin_size).astype(int)

    snvs = snvs[['chrom', 'coord', 'ref', 'alt', 'plot_bin']].drop_duplicates()

    count_table = snvs.groupby(['plot_bin']).size()
    count_table = count_table.reindex(plot_bins).fillna(0)

    ax.bar(plot_bin_starts, count_table.values, width=bin_size)

    if full_genome:
        ax.set_xlim((-0.5, refgenome.info.chromosome_end.max()))
    else:
        ax.set_xlim((min_x, max_x))
    ax.set_xlabel('chrom')
    ax.set_xticks([0] + list(refgenome.info.chromosome_end.values))
    ax.set_xticklabels([])
    ax.set_ylabel('count')
    ax.xaxis.set_minor_locator(matplotlib.ticker.FixedLocator(refgenome.info.chromosome_mid))
    ax.xaxis.set_minor_formatter(matplotlib.ticker.FixedFormatter(refgenome.info.chromosomes))
    ax.xaxis.grid(False, which="minor")



import collections
import numpy as np
from scipy import stats


def calculate_smoothed(position, adjacent, stddev):
    return np.sum(stats.norm(loc=position, scale=stddev).pdf(adjacent))


def calculate_adjacent_density(positions, stddev):
    """ Smoothed position density

    Args:
        positions(numpy.array) : list of positions
        stddev(float) : standard deviation for gaussian smoothing

    Calculate the kernal density of neighbouring positions for each
    position.  Use a gaussian kernal with given standard deviation
    """

    bandwidth = 3.0 * stddev
    left_bandwidth_positions = collections.deque()
    left_counts = list()
    for position in positions.values:
        while len(left_bandwidth_positions) > 0 and left_bandwidth_positions[0] < position - bandwidth:
            left_bandwidth_positions.popleft()
        left_counts.append(calculate_smoothed(position, left_bandwidth_positions, stddev))
        left_bandwidth_positions.append(position)
    right_bandwidth_positions = collections.deque()
    right_counts = list()
    for position in reversed(positions.values):
        while len(right_bandwidth_positions) > 0 and right_bandwidth_positions[0] > position + bandwidth:
            right_bandwidth_positions.popleft()
        right_counts.append(calculate_smoothed(position, right_bandwidth_positions, stddev))
        right_bandwidth_positions.append(position)
    right_counts = list(reversed(right_counts))
    return np.array(left_counts) + np.array(right_counts)


def calculate_adjacent_distance(positions):
    """ Distance to nearest adjacent.

    Args:
        positions(numpy.array) : list of positions

    Returns:
        distances(numpy.array): list of distances

    """
    dist_adj = np.absolute(positions[:-1] - positions[1:])

    dist_left = np.ones(positions.shape) * np.inf
    dist_left[1:] = dist_adj

    dist_right = np.ones(positions.shape) * np.inf
    dist_right[:-1] = dist_adj

    dist = np.minimum(dist_left, dist_right)
    dist[dist == np.inf] = np.nan

    return dist



from setuptools import find_packages, setup
import versioneer


setup(
    name='wgs_analysis',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Whole genome sequencing analysis library',
    packages=find_packages(),
    package_data={'':['data/*']},
)



import numpy as np
import pandas as pd


def cluster_copies(copies):
    """
    Cluster copies by chromosome and copy state
    
    Args:
        copies (pandas.DataFrame): `cnv_site` table

    Returns:
        pandas.DataFrame: table of clustered copy numbers

    Cluster copies by copy state (major, minor, major_sub, minor_sub) and by
    chromosome.  Calculate the average of major_raw and minor_raw weighted
    by length.

    """
    
    copies_table = copies.copy()
    copies_table = copies_table.set_index(['chrom', 'start', 'end', 'site_id'])
    copies_table = copies_table[['major', 'minor', 'major_sub', 'minor_sub']]
    copies_table = copies_table.unstack().fillna(0.0)
    copies_table.columns = [':'.join(a) for a in copies_table.columns]
    
    copies_columns = list(copies_table.columns.values)
    
    copies_group = copies_table.reset_index().drop_duplicates(copies_columns)
    copies_group['copies_group'] = list(xrange(len(copies_group.index)))
    
    copies_table.reset_index(inplace=True)
    copies_table = copies_table.merge(copies_group[copies_columns+['copies_group']], left_on=copies_columns, right_on=copies_columns)
    
    copies = copies.merge(copies_table[['chrom', 'start', 'end', 'copies_group']], left_on=['chrom', 'start', 'end'], right_on=['chrom', 'start', 'end'])
    
    copies['major_raw'] = copies['major_raw'] * copies['length']
    copies['minor_raw'] = copies['minor_raw'] * copies['length']
    
    # Aggregate for grouping by copy group.
    # For each set of copies, aggregate by minimum which has no effect
    # because they are the same across the group
    agg_op = dict({'start':np.min, 'end':np.max, 'length':np.sum, 'major_raw':np.sum, 'minor_raw':np.sum})
    
    # Group and remove copy_group column
    copies = copies.groupby(['site_id', 'chrom', 'copies_group']).agg(agg_op).reset_index()
    
    copies['major_raw'] = copies['major_raw'] / copies['length']
    copies['minor_raw'] = copies['minor_raw'] / copies['length']
    
    return copies


def aggregate_adjacent(cnv, value_cols=(), stable_cols=(), length_normalized_cols=(), summed_cols=()):
    """ Aggregate adjacent segments with similar copy number state.

    see: https://github.com/amcpherson/remixt/blob/master/remixt/segalg.py

    Args:
        cnv (pandas.DataFrame): copy number table

    KwArgs:
        value_cols (list): list of columns to compare for equivalent copy number state
        stable_cols (list): columns for which values are the same between equivalent states
        length_normalized_cols (list): columns that are width normalized for equivalent states

    Returns:
        pandas.DataFrame: copy number with adjacent segments aggregated
    """

    # Group segments with same state
    cnv = cnv.sort_values(['chromosome', 'start'])
    cnv['chromosome_index'] = np.searchsorted(np.unique(cnv['chromosome']), cnv['chromosome'])
    cnv['diff'] = cnv[['chromosome_index'] + value_cols].diff().abs().sum(axis=1)
    cnv['is_diff'] = (cnv['diff'] != 0)
    cnv['cn_group'] = cnv['is_diff'].cumsum()

    def agg_segments(df):
        a = df[stable_cols].iloc[0]

        a['chromosome'] = df['chromosome'].min()
        a['start'] = df['start'].min()
        a['end'] = df['end'].max()
        a['length'] = df['length'].sum()

        for col in length_normalized_cols:
            a[col] = (df[col] * df['length']).sum() / (df['length'].sum() + 1e-16)

        for col in summed_cols:
            a[col] = df[col].sum()

        return a

    aggregated = cnv.groupby('cn_group').apply(agg_segments)

    for col in aggregated:
        aggregated[col] = aggregated[col].astype(cnv[col].dtype)

    return aggregated


def calculate_gene_copy(cnv, genes, cols):
    """ Calculate the copy number segments overlapping each gene

    Args:
        cnv (pandas.DataFrame): copy number table
        genes (pandas.DataFrame): gene table
        cols (list of str): copy number columns

    Returns:
        pandas.DataFrame: segment copy number for each gene

    The input copy number table is assumed to have columns: 
        'chromosome', 'start', 'end', 'width'
    in addition to the columns from cols

    The input genes table is assumed to have columns:
        'chromosome', 'gene_start', 'gene_end', 'gene_name',
        'gene_id'

    The output segment copy number table should have columns:
        'chromosome', 'gene_start', 'gene_end', 'gene_name',
        'gene_id', 'overlap_start', 'overlap_end', 'overlap_width',
    in addition to the columns from cols

    """
    data = []

    for chromosome in cnv['chromosome'].unique():
        chr_cnv = cnv[cnv['chromosome'] == chromosome]
        chr_genes = genes[genes['chromosome'] == chromosome]
        
        # Iterate through segments, calculate overlapping genes
        for idx, row in chr_cnv.iterrows():

            # Subset overlapping genes
            overlapping_genes = chr_genes[
                ~((chr_genes['gene_end'] < row['start']) | (chr_genes['gene_start'] > row['end']))].copy()

            overlapping_genes['overlap_start'] = overlapping_genes['gene_start'].clip(lower=row['start'])
            overlapping_genes['overlap_end'] = overlapping_genes['gene_end'].clip(upper=row['end'])
            overlapping_genes['overlap_width'] = overlapping_genes['overlap_end'] - overlapping_genes['overlap_start'] - 1

            for col in cols:
                overlapping_genes[col] = row[col]

            data.append(overlapping_genes)

    data = pd.concat(data, ignore_index=True)

    return data




'''
# SNV primer design

Design primers for SNVs.  Input to design is a pandas table of positions.  Sequences are created on the fly and
additional single nucleotide changes are masked if available.  Uses the design_primers module to design the primers and
return a table.

'''

import os
import functools
import argparse
import subprocess
import pandas as pd

import design_utils
import design_primers
import blat_server


def get_requirements_filename(experiment_type):
    return os.path.join(os.path.dirname(__file__), '{0}_snv_requirements.tsv'.format(experiment_type))


default_snv_requirements = get_requirements_filename('seqval')


def create_snv_sequence(genome_fasta, chrom, coord, variant_vcfs=[]):
    '''
    Create sequences encompassing a position, and mask variants with N
    '''
    start = max(1, coord - 200)
    end = coord + 200

    sequence = design_utils.create_sequence(genome_fasta, chrom, start, end, variant_vcfs)
    
    sequence = list(sequence)
    sequence[coord-start:coord-start+1] = ['[', 'N', ']']
    sequence = ''.join(sequence)

    return sequence


def create_snv_sequences(genome_fasta, snvs, variant_vcfs=[]):
    '''
    Generate sequences for a table of snvs
    '''
    for idx, row in snvs.iterrows():
        seq_id = row['seq_id']
        sequence = create_snv_sequence(genome_fasta, row['chrom'], row['coord'], variant_vcfs)
        yield seq_id, sequence


def snv_primer_product_check(snvs, primers, products):
    '''
    Ensure the product overlaps the SNV, add product start and end to the primer table
    '''
    snv_products = products[['seq_id', 'primer_id', 'chrom', 'start', 'end']].merge(snvs[['seq_id', 'chrom', 'coord']].drop_duplicates(), on=['seq_id', 'chrom'], how='inner')
    snv_products = snv_products[(snv_products['coord'] > snv_products['start']) & (snv_products['coord'] < snv_products['end'])]
    snv_products = snv_products.rename(columns={'start':'product_start', 'end':'product_end'})
    primers = primers.merge(snv_products[['seq_id', 'primer_id', 'product_start', 'product_end']].drop_duplicates(), on=['seq_id', 'primer_id'], how='inner')
    products = products.merge(snv_products[['seq_id', 'primer_id']].drop_duplicates(), how='inner')
    return primers, products


def design(genome_filename, snvs, variant_vcfs=[], requirements_filename=default_snv_requirements, max_stage=-1, max_primers=-1, primer3_params=None):
    """
    Design primers for an SNV validation experiment

    Args:
        genome_filename(str) : genome file in fasta format
        snvs(pandas.DataFrame) : table of SNV information

    KwArgs:
        variant_vcfs(list) : list of VCF files with germline variants
        requirements_filename(str) : requirements tsv file detailing design requirements by stage
        max_stage(int) : maximum stage before failing, -1 to try all stages
        max_primers(int) : maximum number of primers to design before returning
        primer3_params(dict) : Additional Primer 3 Parameters

    Returns:
        pandas.DataFrame of primer information

    The table of SNVs should have the following required columns:
     * seq_id : unique identifier for the SNV
     * chrom : chromosome of the SNV
     * coord : coordinate of the SNV

    """
    snv_sequences = create_snv_sequences(genome_filename, snvs, variant_vcfs)
    design_snv_callback = functools.partial(snv_primer_product_check, snvs)

    primer_table = design_primers.design(requirements_filename, snv_sequences, design_callback=design_snv_callback, max_stage=max_stage, max_primers=max_primers, primer3_params=primer3_params)

    primer_table = primer_table.merge(snvs, on='seq_id', how='inner')

    return primer_table



if __name__ == '__main__':

    argparser = argparse.ArgumentParser()
    argparser.add_argument('genome', help='genome fasta filename')
    argparser.add_argument('snvs', help='snvs table filename')
    argparser.add_argument('primers', help='primer table filename')
    argparser.add_argument('--requirements', required=False, default=default_snv_requirements, help='design requirements table')
    argparser.add_argument('--varvcfs', nargs='+', required=False, default=[], help='variant vcfs for masking')
    args = argparser.parse_args()

    snvs = pd.read_csv(args.snvs, sep='\t', converters={'chrom':str})[['chrom', 'coord']].drop_duplicates()

    with blat_server.BlatServer([args.genome]):
        primer_table = design(args.genome, snvs, args.varvcfs)

    primer_table.to_csv(args.primers, sep='\t', index=False, na_rep='NA')








'''
Created on Apr 7, 2016

@author: Andrew Roth
'''
import dollo.trees
import gzip
import pandas as pd
import pickle


def load_tree(file_name):
    
    with gzip.GzipFile(file_name, 'r') as fh:
        tree = pickle.load(fh)
    
    return tree


def tree_to_table(tree):
    tree_table = []

    for n in tree.nodes:
        for child in n.children:
            child.parent = n.label
    
    for n in tree.nodes:
        node_attr = {}

        for key, value in n.__dict__.iteritems():
            if isinstance(value, (int, float, str)):
                node_attr[key] = value

        tree_table.append(node_attr)
        
    return pd.DataFrame(tree_table)


def table_to_tree(table):
    
    table = table.sort_values(by='parent')

    nodes = {}
    
    skip = table.isnull()
    
    for idx in table.index:
        n = dollo.trees.TreeNode()
            
        for col in table.columns:
            if not skip.loc[idx, col]:
                setattr(n, col, table.loc[idx, col])
                
        nodes[n.label] = n
    
    for label, node in nodes.iteritems():
        if hasattr(node, 'parent') and node.parent >= 0:
            nodes[node.parent].children.append(node)

    return nodes[0]


'''
# Color sets

Standardized color sets for consistent plots.

'''
from collections import OrderedDict

import brewer2mpl
import collections
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sb

import wgs_analysis.plots as plots
import wgs_analysis.refgenome as refgenome


max_color_idx = 12

def get_color_set(labels):

    cmap = plt.get_cmap('Dark2')

    color_set = dict()
    for idx, label in enumerate(labels):
        assert idx <= max_color_idx
        color_set[label] = cmap(float(idx) / float(max_color_idx))

    return color_set


def create_chromosome_color_map():

    colors = sb.color_palette("muted")
    color_set = [colors[i % 2] for i in range(len(refgenome.info.chromosomes))]
    chromosome_color = collections.OrderedDict(zip(refgenome.info.chromosomes, color_set))
    return chromosome_color


def get_single_cell_color_set():
    snv_palette = brewer2mpl.get_map('YlGnBu', 'Sequential', 3).mpl_colors

    colors = OrderedDict((('A', snv_palette[0]),
                          ('AB', snv_palette[1]),
                          ('B', snv_palette[2]),
                          ('Unknown', 'gray'),
                          ('Absent', '#f0f0f0'),
                          ('Present', '#636363')))
    
    return colors

def get_data_frame_color_set(df, map_series='Set1', map_type='Qualitative'):
    '''
    Return a dictionary mapping values in data frame to colors in a color brewer set.
    '''
    values = sorted(pd.unique(df.values.ravel()))
    
    palette = sb.color_palette(map_series, len(values))

    colors = OrderedDict(zip(values, palette))
    
    return colors

def get_tree_node_color_set(patient_id, is_site_tree=True):
    if is_site_tree:
        tree = helpers.get_snv_tree(patient_id)
    
    else:
        tree = helpers.get_single_cell_tree(patient_id)
    
    node_labels = sorted([node.label for node in tree.nodes])

    color_set = get_color_set(node_labels)
    
    return OrderedDict(zip(node_labels, [color_set[x] for x in node_labels]))

def get_loss_deleted_colors():
    c = get_loss_deleted_color_map()
    
    return [c['Unexplained'], c['Deleted']]

def get_loss_deleted_color_map():
    return {'Unexplained' :'#1b9e77', 'Deleted' : '#d95f02'}

def add_loss_deleted_legend(ax, frame=True):
    color_map = get_loss_deleted_color_map()
    
    palette = color_map.values()

    artists = [plt.Circle((0, 0), color=c) for c in palette]
    
    legend = ax.legend(artists, color_map.keys(), loc='upper right', frameon=frame)

    if frame:
        frame = legend.get_frame()
        frame.set_facecolor('white')
        frame.set_edgecolor('white')

    
    



import itertools
import pandas as pd
import numpy as np

def values_to_fixed_width(values, widths, fill=' ', sep='|'):
    text = []
    text.append(sep)
    for v, w in zip(values, widths):
        text.append(fill)
        text.append(v)
        text.append(fill * (w - len(v)))
        text.append(fill + sep)
    return ''.join(text)


def df_to_markdown(df):
    ''' Convert a DataFrame to a markdown table
    '''

    markdown_text = []

    df = df.applymap(lambda v: '{0}'.format(v))
    cols = ['`{0}`'.format(a) for a in df.columns]

    data_width = df.applymap(lambda v: len(v)).max(axis=0).values
    header_width = np.array([len(str(a)) for a in cols])

    cell_width = np.maximum(data_width, header_width)

    markdown_text.append(values_to_fixed_width(cols, cell_width))

    markdown_text.append(values_to_fixed_width(itertools.repeat(''), cell_width, fill='-'))

    for idx, row in df.iterrows():
        markdown_text.append(values_to_fixed_width(row.values, cell_width))

    markdown_text = '\n'.join(markdown_text)

    return markdown_text
